import json
from datetime import timedelta

from django.core.exceptions import ImproperlyConfigured
from django.db import transaction
from django.tasks import Task
from django.tasks import TaskResult
from django.tasks import TaskResultStatus
from django.tasks.backends.base import BaseTaskBackend
from django.tasks.signals import task_enqueued
from django.utils import timezone
from django.utils.crypto import get_random_string
from django.utils.module_loading import import_string
from google import auth
from google.cloud.tasks_v2 import CloudTasksClient
from google.cloud.tasks_v2 import HttpMethod


class CloudTasksBackend(BaseTaskBackend):
    supports_defer = True

    # Cloud Tasks default results APIs are not ideal for most uses. For example, once a task is complete, requesting the
    # status of the task will return a 404, rather than a success message. This means there is really no way to know if
    # it finished successfully, failed, or didn't really enqueue properly.
    # Supporting results is outside the scope of this backend, and can be built as an extension using a Django Model
    # or key-value store.
    supports_get_result = False

    # Google Cloud Tasks does not support priority, so this backend does not support priority.
    supports_priority = False

    def __init__(self, alias, params):
        super().__init__(alias, params)

        self.client = None

    def get_view_authn(self):
        authn_class = self.options.get("VIEW_AUTHN", -1)
        if authn_class == -1:
            raise ImproperlyConfigured(
                "The view authentication class must be specified. To turn off authentication, set VIEW_AUTHN to None."
            )

        if authn_class is None:
            return None

        authn_class = import_string(authn_class)

        return authn_class(**self.options.get("VIEW_AUTHN_PARAMS"))

    def get_project_id(self):
        if not self.options.get("PROJECT_ID"):
            raise ImproperlyConfigured("The project ID must be specified.")

        return self.options.get("PROJECT_ID")

    def get_location(self):
        if not self.options.get("LOCATION"):
            raise ImproperlyConfigured("The location must be specified.")

        return self.options.get("LOCATION")

    def get_credentials(self):
        if hasattr(self, "_credentials"):
            return self._credentials

        configured_credentials = self.options.get("CREDENTIALS")
        if configured_credentials:
            self._credentials = configured_credentials
        else:
            self._credentials, _ = auth.default(scopes=["https://www.googleapis.com/auth/cloud-platform"])

        if self._credentials is None:
            raise ImproperlyConfigured("The credentials must be specified or configured via the machine.")

        return self._credentials

    def get_default_target(self):
        target = self.options.get("DEFAULT_TARGET")

        if target is None:
            raise ImproperlyConfigured("The default target must be specified.")

        return target

    def get_client(self):
        if not self.client:
            self.client = CloudTasksClient(credentials=self.get_credentials())

        return self.client

    def get_parent_path(self, queue_name: str) -> str:
        return self.get_client().queue_path(self.get_project_id(), self.get_location(), queue_name)

    def get_task_path(self, queue_name: str, task_id: str) -> str:
        return self.get_client().task_path(self.get_project_id(), self.get_location(), queue_name, task_id)

    def get_enqueue_on_commit(self, task: Task):
        return self.options.get("ENQUEUE_ON_COMMIT", False)

    def enqueue(self, task: Task, args, kwargs):
        self.validate_task(task)

        # Create the task
        task_result = TaskResult(
            task=task,
            id=get_random_string(32),
            status=TaskResultStatus.READY,
            enqueued_at=None,
            started_at=None,
            last_attempted_at=None,
            finished_at=None,
            args=args,
            kwargs=kwargs,
            backend=self.alias,
            errors=[],
            worker_ids=[],
        )

        task_data = {
            "task_path": task.module_path,
            "args": args,
            "kwargs": kwargs,
        }

        cloud_task = {
            "http_request": {
                "http_method": HttpMethod.POST,
                "url": self.get_default_target(),
                "headers": {"Content-type": "application/json"},
                "body": json.dumps(task_data).encode(),
            },
            "name": self.get_task_path(task.queue_name, task_result.id),
        }

        def _enqueue():
            nonlocal cloud_task

            if task.run_after:
                run_after = task.run_after
                if isinstance(task.run_after, timedelta):
                    run_after = timezone.now() + task.run_after

                cloud_task["schedule_time"] = run_after

            self.get_client().create_task(
                request={
                    "parent": self.get_parent_path(task.queue_name),
                    "task": cloud_task,
                }
            )

            object.__setattr__(task_result, "enqueued_at", timezone.now())
            task_enqueued.send(type(self), task_result=task_result)

        if self.get_enqueue_on_commit(task):
            transaction.on_commit(_enqueue)
        else:
            _enqueue()

        return task_result



import json
from typing import TypedDict

from django.contrib.auth.models import AnonymousUser
from django.core.exceptions import ImproperlyConfigured
from django.core.exceptions import SuspiciousOperation
from django.http import HttpRequest
from django.http import JsonResponse
from django.tasks import Task
from django.tasks import TaskContext
from django.tasks import TaskResult
from django.tasks import TaskResultStatus
from django.tasks import task_backends
from django.tasks.base import TaskError
from django.tasks.signals import task_finished
from django.tasks.signals import task_started
from django.utils import timezone
from django.utils.module_loading import import_string
from django.views import View

from django_tasks_gcp.authn import ViewAuth
from django_tasks_gcp.backend import CloudTasksBackend
from django_tasks_gcp.results import CloudTaskResult
from django_tasks_gcp.utils import get_exception_traceback
from django_tasks_gcp.utils import get_module_path


class Input(TypedDict):
    task_path: str
    args: list[str]
    kwargs: dict[str, str]


class TaskView(View):
    backend_name: str | None = None

    def __init__(self, **kwargs):
        self.backend_name = kwargs.pop("backend_name", None)

        super().__init__(**kwargs)

    def post(self, request):
        identity = self.authenticate(request)

        if identity is None:
            self.fail_authentication()
            return JsonResponse({"success": False}, status=401)

        data = self.parse_content(request)

        data = self.validate_input(data)

        task = self.get_task(data)

        task_result = self.get_task_result(request, data, task)

        self.run_task(task_result, data)

        if task_result.status == TaskResultStatus.FAILED:
            return JsonResponse({"success": False}, status=400)

        return JsonResponse({"success": True})

    def authenticate(self, request: HttpRequest):
        """
        Authenticate the request.

        This authentication is done without parsing the task first. This is purposeful since we are working with an
        endpoint that could be open to the world. By only using configurations on the application, we help avoid DDoS
        attacks passing in large JSON payloads.

        If the view was configured with a backend explicitly, then that will
        be used for authentication. Otherwise, the first CloudTasksBackend found in the list of configured backends will
        be used instead.

        To configure an explicit backend, register the view was `TaskView.as_view(backend_name="my_backend")`.
        """
        backend = None
        if self.backend_name:
            backend = task_backends[self.backend_name]

            if backend is None:
                raise ImproperlyConfigured(f"Task backend '{self.backend_name}' is not configured.")
        else:
            for task_backend in task_backends.all():
                if isinstance(task_backend, CloudTasksBackend):
                    backend = task_backend
                    break

            if backend is None:
                raise ImproperlyConfigured("No CloudTasksBackend is configured.")

        authn = backend.get_view_authn()

        if authn is None:
            return AnonymousUser()

        if not isinstance(authn, ViewAuth):
            raise ImproperlyConfigured("View authentication must be a subclass of ViewAuth.")

        return authn.authenticate(request)

    def fail_authentication(self):
        pass

    def parse_content(self, request: HttpRequest) -> dict:
        return json.loads(request.body)

    def validate_input(self, data: dict) -> Input:
        if "task_path" not in data:
            raise ValueError("Missing task_path in request body.")

        if not isinstance(data["task_path"], str):
            raise ValueError("task_path must be a string.")

        if "args" not in data:
            raise ValueError("Missing args in request body.")

        if not isinstance(data["args"], list):
            raise ValueError("args must be a list.")

        if "kwargs" not in data:
            raise ValueError("Missing kwargs in request body.")

        if not isinstance(data["kwargs"], dict):
            raise ValueError("kwargs must be a dict.")

        return Input(**data)

    def get_task(self, data: Input) -> Task:
        try:
            task = import_string(data["task_path"])

            if not isinstance(task, Task):
                raise SuspiciousOperation(f"{data['task_path']} is not a valid task.")

            return task
        except ImportError:
            raise SuspiciousOperation(f"Task {data['task_path']} does not exist.")

    def get_task_result(self, request: HttpRequest, data: Input, task: Task):
        task_id = request.headers.get("X-Cloudtasks-Taskname")
        retry_count = int(request.headers.get("X-CloudTasks-TaskRetryCount", 0))  # 0 for the first attempt

        return CloudTaskResult(
            task=task,
            id=task_id,
            status=TaskResultStatus.RUNNING,
            started_at=timezone.now(),
            finished_at=None,
            backend=task.get_backend(),
            args=data.get("args", []),
            kwargs=data.get("kwargs", {}),
            retry_count=retry_count,
            # The following arguments are not supported by this library
            enqueued_at=None,
            last_attempted_at=None,
            errors=[],
            worker_ids=[],
        )

    def get_task_context(self, task_result: TaskResult) -> TaskContext:
        return TaskContext(task_result=task_result)

    def run_task(self, task_result: TaskResult, data: Input):
        task = task_result.task

        task_started.send(sender=task.get_backend(), task_result=task_result)

        try:
            if task.takes_context:
                return_value = task.call(self.get_task_context(task_result), *data["args"], **data["kwargs"])
            else:
                return_value = task.call(*data["args"], **data["kwargs"])

            object.__setattr__(task_result, "_return_value", return_value)
            object.__setattr__(task_result, "status", TaskResultStatus.SUCCESSFUL)
            object.__setattr__(task_result, "finished_at", timezone.now())

            task_finished.send(sender=task.get_backend(), task_result=task_result)
        except BaseException as e:
            task_result.errors.append(
                TaskError(
                    exception_class_path=get_module_path(type(e)),
                    traceback=get_exception_traceback(e),
                )
            )

            object.__setattr__(task_result, "finished_at", timezone.now())
            object.__setattr__(task_result, "status", TaskResultStatus.FAILED)

            task_finished.send(sender=task.get_backend(), task_result=task_result)



from dataclasses import dataclass

from django.tasks import TaskResult


@dataclass(frozen=True, slots=True, kw_only=True)
class CloudTaskResult(TaskResult):
    retry_count: int

    @property
    def attempts(self):
        return self.retry_count + 1



import abc
from typing import Any

from django.http import HttpRequest
from google.auth.transport import requests
from google.oauth2 import id_token


class ViewAuth(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def authenticate(self, request: HttpRequest) -> Any | None:
        pass


class OIDCTokenAuth(ViewAuth):
    def __init__(self, service_account_email: str | None):
        self.service_account_email = service_account_email

    def authenticate(self, request: HttpRequest) -> Any | None:
        token = request.headers.get("Authorization")
        if not token:
            return None

        token = token.split(" ")
        if len(token) != 2 or token[0] != "Bearer":
            return None

        token = token[1]

        try:
            idinfo = id_token.verify_oauth2_token(token, requests.Request())

            if self.service_account_email and idinfo["email"] != self.service_account_email:
                return None

            return idinfo
        except BaseException:
            return None



from traceback import format_exception
from typing import Any


def get_module_path(val: Any) -> str:
    return f"{val.__module__}.{val.__qualname__}"


def get_exception_traceback(exc: BaseException) -> str:
    return "".join(format_exception(exc))
